IMPLEMENTATION MODULE AnalyzeProcs;

  (*

        Implementation and Revisions:
        ============================

        Author  Date        Description
        ------  ----        -----------

        dg      11.12.89    First implementation
        af		21.12.89	ShowWaitSymbol and HideWaitSymbol
        					now used to bracket waiting by
        					the user during analysis

  *)

  FROM DMMaster IMPORT DialogMachineTask, ShowWaitSymbol, HideWaitSymbol;
  FROM DMFiles IMPORT  Response ;
  
  FROM ClientsAux IMPORT 
    outF, outFOpen, writeOnFile,  Write, WriteString, WriteLn, WriteCardinal, 
    WriteInteger, CompareStr, killed ;
  
  FROM IDTables IMPORT
    MaxIDs, IDString, MaxIDLength, IDTablesSymbol, IDRecord,  ShowIDTable,
    ShowResIDTable,AddID, LookupID, CountUpForThisID, GetFirstIDRecord,
    GetNextIDRecord, WriteIDType, WriteLnAndCheck, ResetIDCountsOfAllRecords,
    OccurenceType ;

  FROM GetSym IMPORT
    InitGetSym, NextSymbol, LineNumber, GetCurrentID, copyFileToOutput;


  TYPE
     Error  = (FileNotFound, NoDEFModule, NoMODModule, OwnMODModule,
               IDNotExported, JustModuleName, SyntaxError, StructuredVariable,
               UnknownQualifier, ProcedureTwice, IDTableFull );
               
  VAR 
    CurrentSymbol : IDTablesSymbol ;
    CurrentID     : IDString ;
    defModuleName,
    clientModuleName: IDString ;
    IDCount       : CARDINAL ;
    ProcedureCount: CARDINAL ;
    anImportOccures : BOOLEAN ;
    
    
  (*. ****************************************** .*)
  (*. Error handling                             .*)
  (*. ****************************************** .*) 
  
  PROCEDURE DisplayError( ErrorCondition: Error; String: ARRAY OF CHAR);
  BEGIN
     IF NOT killed THEN
       WriteLn;
       WriteString( ">>>   ERROR: " );
       CASE ErrorCondition OF
          FileNotFound:
                WriteString( "The file '" ); WriteString( String );
                WriteString( "' could not be found." ); WriteLn; |
          UnknownQualifier:
                WriteString( "Unknown qualifier:  /" ); WriteString( String );
                WriteLn; |
          NoDEFModule: 
                WriteString( "The file '" ); WriteString( String ); Write( "'" );
                WriteString( "' contains no DEFINITION MODULE" ); WriteLn;
                WriteString( ">>>   or no exports occure." ); WriteLn; |
          NoMODModule:
                WriteString( "The file '" ); WriteString( String ); Write( "'" );
                WriteString( " contains no IMPLEMENTATION MODULE" ); WriteLn; |
          OwnMODModule:
                WriteString( "The own IMPLEMENTATION MODULE is excepted" ); WriteLn;
                WriteString( ">>>   from analysis." ); WriteLn; |
          IDNotExported: 
                WriteString( "The ident '" ); WriteString( String ); Write( "'" ); WriteLn;
                WriteString( ">>>   is not exported" ); WriteLn; |
          ProcedureTwice: 
                WriteString( "The procedure '" ); WriteString( String ); Write( "'" );
                WriteString( " is defined more then once." ); WriteLn; |
          SyntaxError:
                WriteString( "Syntax-error:  <" ); WriteString( String );
                WriteString( "> on line " );
                WriteCardinal( LineNumber(), 4 ); WriteLn; |
          StructuredVariable:
                WriteString(  "Structured variable encountered (line " );
                WriteCardinal( LineNumber(), 4 ); WriteString( ")." ); WriteLn; |
          JustModuleName:
                WriteString( "The ident '" ); WriteString( String );
                WriteString( "' preceeds neither IMPORT" ); WriteLn;
                WriteString( ">>>  nor '.'     (line " );
                WriteCardinal( LineNumber(), 4 ); WriteString( ")." ); WriteLn; |
          IDTableFull:
                WriteString( "Idents-table too small  (max." );
                WriteCardinal( MaxIDs, 4 ); WriteString( " entries possible)" ); WriteLn;
       END;
       WriteLn;
     END;(*IF*)
  END DisplayError;


  (*. ****************************************** .*)
  (*. Auxiliary procs needed in module           .*)
  (*. ****************************************** .*) 

  PROCEDURE LookupAndAddIfNotThere( ID: IDString; IDType: IDTablesSymbol );
  VAR
     Found, ok:     BOOLEAN;
     PlaceToInsert: CARDINAL;
  BEGIN
     LookupID( ID, PlaceToInsert, Found );
     IF NOT Found THEN
        AddID( ID, PlaceToInsert, IDType, LineNumber(), ok );
        IF NOT ok THEN
           DisplayError( IDTableFull, "" );
        END;
        INC( IDCount );
     END;
  END LookupAndAddIfNotThere;


  PROCEDURE ReadCLIENTFile( AnalyzeProcedure: PROC );
    VAR clientType: (def, impl, mod) ;
  BEGIN
    clientType := mod ;
    WriteString( "Reading CLIENT-file '" ); WriteString( cliF.filename );
    WriteString( "'..." ); ShowWaitSymbol;
    InitGetSym( cliF );
    CurrentSymbol := NextSymbol();
    IF (CurrentSymbol=ImplementationID) THEN
       CurrentSymbol := NextSymbol();
       clientType := impl ;
    ELSIF (CurrentSymbol=DefinitionID) THEN
       CurrentSymbol := NextSymbol();
       clientType := def ;
    END;
    IF CurrentSymbol = ModuleID THEN
       CurrentSymbol := NextSymbol();
       GetCurrentID( CurrentID );
       IF CompareStr( CurrentID, defModuleName ) = 0 THEN
          WriteLn ;
          DisplayError( OwnMODModule, "" );
       ELSE
          CASE clientType OF
            def   : WriteString( "   >> DEFINITION MODULE " );
          | impl  : WriteString( "   >> IMPLEMENTATION MODULE " );
          | mod   : WriteString( "   >> MODULE " );
          END(*CASE*);
          clientModuleName := CurrentID ;
          WriteString( CurrentID );
          WriteLn;
          CurrentSymbol := NextSymbol();
          AnalyzeProcedure;
          HideWaitSymbol;
       END;
    ELSE
       DisplayError( NoMODModule, cliF.filename );
    END;
  END ReadCLIENTFile;
  

  (*. ****************************************** .*)
  (*. Procedures analyzing client-module         .*)
  (*. ****************************************** .*) 
  
  PROCEDURE AnalyzeImportingFile;
    VAR
      QualifiedImport:   BOOLEAN;
      UnqualifiedImport: BOOLEAN;
      ok:                BOOLEAN;
  BEGIN
     QualifiedImport :=   FALSE;
     UnqualifiedImport := FALSE;
     WHILE ( (CurrentSymbol <> EndSymbol) AND (NOT killed) ) DO
        IF CurrentSymbol = ImportID THEN
           IF NextSymbol() = Identifier THEN
              GetCurrentID( CurrentID );
              IF CompareStr( CurrentID, defModuleName) = 0 THEN
                 QualifiedImport := TRUE;
              END;
           END;
           CurrentSymbol := NextSymbol();
        ELSIF CurrentSymbol = Identifier THEN
             GetCurrentID( CurrentID );
             IF CompareStr( CurrentID, defModuleName) = 0 THEN
                CurrentSymbol := NextSymbol();
                IF CurrentSymbol = ImportID THEN (* scan the IMPORT list *)
                   UnqualifiedImport := TRUE;
                   REPEAT
                      CurrentSymbol := NextSymbol();
                      IF CurrentSymbol = Identifier THEN
                        GetCurrentID( CurrentID );
                        CountUpForThisID( CurrentID, inImportList, ok );
                        IF NOT ok THEN
                           DisplayError( IDNotExported, CurrentID );
                        END;(*IF*)
                      END;(*IF*)
                   UNTIL CurrentSymbol = Semicolon ;
                ELSIF CurrentSymbol = Dot THEN (* it's a qualified import *)
                   IF NextSymbol() = Identifier THEN
                      GetCurrentID( CurrentID );
                      CountUpForThisID( CurrentID, usedQualified, ok );
                      IF NOT ok THEN
                         DisplayError( IDNotExported, CurrentID );
                      END;
                      CurrentSymbol := NextSymbol();
                   END;
                ELSE
                   DisplayError( JustModuleName, CurrentID );
                   CurrentSymbol := NextSymbol();
                END;
             ELSE
                CountUpForThisID( CurrentID, otherOccurence, ok );
                CurrentSymbol := NextSymbol();
             END;
        ELSE
           CurrentSymbol := NextSymbol();
        END;
     END;(*WHILE*)
     IF QualifiedImport OR UnqualifiedImport THEN
        anImportOccures := TRUE ;
     ELSE
        anImportOccures := FALSE ;
     END;
  END AnalyzeImportingFile;


  PROCEDURE CollectIDs;
  VAR
     CurrentSection: (Beginning, ConstSection, TypeSection, VarSection,
                      ProcedureSection);

     PROCEDURE SkipParameterList;
     (*
      *  If in a variable or type declaration we find an ID of type PROCEDURE
      *  we must skip the parameterlist because of the possible VAR in it
      *)
     BEGIN
        CurrentSymbol := NextSymbol();
        IF CurrentSymbol = OpenParen THEN
           REPEAT
              CurrentSymbol := NextSymbol();
           UNTIL CurrentSymbol = CloseParen;
           CurrentSymbol := NextSymbol();
           IF CurrentSymbol = Colon THEN       (* it's a function procedure *)
              (* skip colon and function return type *)
              CurrentSymbol := NextSymbol();
              CurrentSymbol := NextSymbol();
           END;
        END;
     END SkipParameterList;

  BEGIN
     CurrentSection := Beginning;
     WHILE ((CurrentSymbol <> EndSymbol) AND (NOT killed)) DO
        IF CurrentSymbol = ConstID THEN
           CurrentSection := ConstSection;
           CurrentSymbol := NextSymbol();
        ELSIF CurrentSymbol = TypeID THEN
           CurrentSection := TypeSection;
           CurrentSymbol := NextSymbol();
        ELSIF CurrentSymbol = VarID THEN
           CurrentSection := VarSection;
           CurrentSymbol := NextSymbol();
           WHILE CurrentSymbol = Identifier DO
              (* enter the variable name *)
              GetCurrentID( CurrentID );
              LookupAndAddIfNotThere( CurrentID, VarID );
              CurrentSymbol := NextSymbol();
              WHILE CurrentSymbol = Comma DO (* process ID list *)
                 CurrentSymbol := NextSymbol();
                 IF CurrentSymbol = Identifier THEN
                    GetCurrentID( CurrentID );
                    LookupAndAddIfNotThere( CurrentID, VarID );
                 ELSE
                    DisplayError( SyntaxError, "no variable ID" );
                 END;
                 CurrentSymbol := NextSymbol();  (* either comma or colon *)
              END;
              CurrentSymbol := NextSymbol(); (* skip the colon and get the type ID *)
              IF CurrentSymbol = ProcedureID THEN
                 SkipParameterList;
              ELSIF CurrentSymbol = Identifier THEN
                 CurrentSymbol := NextSymbol(); (* skip type ID *)
              ELSE
                 DisplayError( StructuredVariable, "" );
                 REPEAT
                    CurrentSymbol := NextSymbol();
                 UNTIL CurrentSymbol = Semicolon;
              END;
              CurrentSymbol := NextSymbol(); (* read the semicolon *)
           END;
        ELSIF CurrentSymbol = ProcedureID THEN
           CurrentSection := ProcedureSection;
           CurrentSymbol := NextSymbol();
           IF CurrentSymbol = Identifier THEN
              (* enter the procedure name *)
              GetCurrentID( CurrentID );
              LookupAndAddIfNotThere( CurrentID, ProcedureID );
              SkipParameterList;
           ELSE
              DisplayError( SyntaxError, "no procedure ID" );
           END;
        ELSIF CurrentSymbol = EqualSymbol THEN
           (* the identifier BEFORE the "=" must be entered        *)
           (* this takes care of all CONST and TYPE declarations   *)
           GetCurrentID( CurrentID );
           CurrentSymbol := NextSymbol(); (* skip the '=' *)
           IF CurrentSection = ConstSection THEN
              LookupAndAddIfNotThere( CurrentID, ConstID );
              CurrentSymbol := NextSymbol();
           ELSE
              LookupAndAddIfNotThere( CurrentID, TypeID );
              IF CurrentSymbol = ProcedureID THEN
                 SkipParameterList;
              ELSE
                 CurrentSymbol := NextSymbol();
              END;
           END;
        ELSE
           (* just skip all the rest of them *)
           CurrentSymbol := NextSymbol();
        END;
     END;
  END CollectIDs;
  
  
  PROCEDURE CountProcedureIDs;
  VAR
     PCount:        CARDINAL;
     Found, ok:     BOOLEAN;
     PlaceToInsert: CARDINAL;
     i:             CARDINAL;
  BEGIN
     PCount := 0;
     WHILE ( (CurrentSymbol <> EndSymbol) AND (NOT killed) ) DO
        IF CurrentSymbol = ProcedureID THEN
           CurrentSymbol := NextSymbol();
           IF CurrentSymbol = Identifier THEN
              INC( PCount );
              GetCurrentID( CurrentID );
              LookupID( CurrentID, PlaceToInsert, Found );
              IF NOT Found THEN
                 AddID( CurrentID, PlaceToInsert, ProcedureID, LineNumber(), ok );
                 IF NOT ok THEN
                    DisplayError( IDTableFull, "" );
                 END;
              ELSE
                 DisplayError( ProcedureTwice, CurrentID );
              END;
           END;
        END;
        CurrentSymbol := NextSymbol();
     END;
     IF NOT killed THEN
       WriteCardinal( PCount, 6 ); WriteString( "  Procedures" );
     END;(*IF*)
     ProcedureCount := ProcedureCount + PCount
  END CountProcedureIDs;


  (*. ****************************************** .*)
  (*. Exported procedures                        .*)
  (*. ****************************************** .*) 
  
  PROCEDURE CheckReferences (newDefFile, 
  							 displayNotOccuringIDs: BOOLEAN;
  							 VAR importOccures: BOOLEAN ) ;  
    VAR
       IDR:               IDRecord;
       IsNotLastRecord:   BOOLEAN;
       occuringIDsCount:   CARDINAL;
  
    PROCEDURE ReadDEFFile  ;
    (* collects all exported IDs *) 
    BEGIN
       WriteString( "Reading DEF-file '" ); WriteString( defF.filename );
       WriteString( "'..." );
       InitGetSym( defF );
       IF NextSymbol() = DefinitionID THEN
          IF NextSymbol() = ModuleID THEN
             IF NextSymbol() = Identifier THEN
                GetCurrentID( defModuleName );
                WriteString( "   >> DEFINITION MODULE " );
                WriteString( defModuleName );
                CurrentSymbol := NextSymbol();
                CollectIDs;
             END;
          END;
       END;
       WriteLn;
    END ReadDEFFile;
    
    PROCEDURE ShowResults;
        
      PROCEDURE CountOccuringIDs ;
      BEGIN
        GetFirstIDRecord( IDR, IsNotLastRecord );
        WHILE IsNotLastRecord DO
           IF (IDR.strOccur > 0) THEN
             INC( occuringIDsCount )
           END;(*IF*)
           GetNextIDRecord( IDR, IsNotLastRecord );
        END;(*WHILE*)
      END CountOccuringIDs ;

    
      PROCEDURE ShowOccuringIDs;
      BEGIN
        WriteString( "Occuring: " ); WriteCardinal( occuringIDsCount, 1 );
        WriteString( " of " ); WriteCardinal( IDCount, 1 );
        WriteString( " idents exported by " ); WriteString( defModuleName ); WriteLn;         
        IF (occuringIDsCount>0) THEN
          WriteLn;
          WriteString( "IDENT                    ObjectNeeded  #IdentOccures    TYPE     DEF-LineNr" );       
          WriteLn;
          GetFirstIDRecord( IDR, IsNotLastRecord );
          WHILE IsNotLastRecord AND (NOT killed) DO
             DialogMachineTask ;
             IF (IDR.strOccur > 0) THEN
                WriteString( IDR.String );WriteString("      ");
                IF (IDR.qualifUse) THEN
                  (*. object is used qualified, e.g. "Module.obj"  .*)
                  Write( "Y" );
                ELSE
                  (*. object is not used qualified .*)
                  IF (IDR.strOccur = 1) THEN
                    (*. if object is imported, the occurance is in the import list,
                    if it is not imported, it must be from an other module.
                    In both cases it is not needed for sure, if it is not a type .*)
                    IF IDR.TypeOfID=TypeID THEN
                      Write( "?" );
                    ELSE
                      Write( "N" );
                    END;(*IF*)
                  ELSE
                    (*. ident-string occures more then once .*)
                    IF IDR.imported THEN
                      (*. Object not used qualified, but imported and occuring more
                      then once: not possible to decide, because occurences of the
                      type "OtherModule.obj" are also counted .*)
                      Write( "?" );
                    ELSE
                      (*. not used qualified, not imported, occuring more then once:
                      occuring ident must refer to object from other module .*)
                      Write( "N" );
                    END;(*IF*)
                  END;(*IF*)
                END;(*IF*)
                WriteCardinal( IDR.strOccur, 14 ); 
                WriteString( "        " );
                WriteIDType( IDR.TypeOfID );
                WriteCardinal( IDR.LineNo, 8 );
                WriteLnAndCheck;
             END;(*IF*)
             GetNextIDRecord( IDR, IsNotLastRecord );
          END;(*WHILE*)
        END;(*IF*)
      END ShowOccuringIDs;
      
      PROCEDURE ShowOtherIDs;
      BEGIN
        WriteLn;
        WriteString("---------------------------------------------------------------------------") ;
        WriteLn; WriteLn;
        WriteString( "Not occuring: "); WriteCardinal( IDCount-occuringIDsCount, 1 );
        WriteString( " of " ); WriteCardinal( IDCount, 1 );
        WriteString( " idents exported by " ); WriteString( defModuleName ); WriteLn;
        IF ((IDCount-occuringIDsCount) > 0) THEN
          WriteLn ;
          WriteString( "IDENT                        TYPE     DEF-LineNr" ); WriteLn;
          GetFirstIDRecord( IDR, IsNotLastRecord );
          WHILE IsNotLastRecord AND (NOT killed) DO
             DialogMachineTask ;
             IF (IDR.strOccur = 0) THEN
                WriteString( IDR.String ); WriteString( "  " );
                WriteIDType( IDR.TypeOfID );
                WriteCardinal( IDR.LineNo, 8 );
                WriteLnAndCheck;
             END;(*IF*)
             GetNextIDRecord( IDR, IsNotLastRecord );
          END;(*WHILE*)
        END;(*IF*)        
      END ShowOtherIDs;
      
    BEGIN
      occuringIDsCount := 0 ;
      WriteLn;
      IF anImportOccures THEN
        WriteString( "   IMPORTs (FROM) " ); WriteString( defModuleName ); WriteLn;        
        WriteLn; 
        CountOccuringIDs ;
        ShowOccuringIDs ;
        IF displayNotOccuringIDs THEN ShowOtherIDs END;(*IF*)
      ELSE
        WriteString( "   No imports from " ); WriteString( defModuleName ); WriteLn;         
      END;(*IF*)
      WriteLn;
    END ShowResults;
    
  BEGIN
    anImportOccures := FALSE ;
    IF newDefFile THEN
      IDCount := 0 ;
      ReadDEFFile ;
    ELSE
      ResetIDCountsOfAllRecords; 
    END;(*IF*)
    IF IDCount > 0 THEN
      IF NOT killed THEN
        ReadCLIENTFile( AnalyzeImportingFile );
        IF NOT killed THEN ShowResults END;(*IF*)
        importOccures := anImportOccures ;
      END;(*IF*)
    ELSE
      DisplayError( NoDEFModule, defF.filename );
      importOccures := FALSE;
    END;(*IF*)
  END CheckReferences;
  

  PROCEDURE CountIDLengths;
  VAR
     Length:          CARDINAL;
     SumOfLength:     CARDINAL;
     i:               CARDINAL;
     IDR:             IDRecord;
     IsNotLastRecord: BOOLEAN;
  BEGIN
     IDCount := 0;      
     SumOfLength := 0;
     i := 0;
     ReadCLIENTFile( CollectIDs );
     IF NOT killed THEN
       GetFirstIDRecord( IDR, IsNotLastRecord );
       WHILE IsNotLastRecord DO
          Length := 0;
          WHILE (Length <= MaxIDLength) AND (IDR.String[ Length ] <> " ") DO
             INC( Length );
          END;
          SumOfLength := SumOfLength + Length;
          INC( i );
          WriteCardinal( i, 4 ); WriteString("   " );
          WriteString( IDR.String );
          WriteCardinal( Length, 6 );
          WriteLnAndCheck;
          GetNextIDRecord( IDR, IsNotLastRecord );
       END;
       IF IDCount>0 THEN
         WriteString( "Average length of identifiers: " );
         WriteCardinal( SumOfLength DIV IDCount, 4 ); WriteLn;
       END;(*IF*)
       WriteLn;
     END;(*IF*)
  END CountIDLengths;


  PROCEDURE CountProcedures;
  BEGIN
     IDCount := 0;      
     ProcedureCount := 0;
     ReadCLIENTFile( CountProcedureIDs );
     IF NOT killed THEN
       ShowIDTable;
       WriteString( "Total " ); WriteCardinal( ProcedureCount, 4 );
       WriteString( " procedures" );
       IF ProcedureCount > IDCount THEN
          WriteString( " (included multiple defined)" );
       END;
       Write( "." ); WriteLn;
     END;(*IF*)
  END CountProcedures;


BEGIN
  defF.res := notdone;
  cliF.res := notdone;
END AnalyzeProcs.
