MODULE Clients ;
(*
 *  Takes one .DEF file and one or more .MOD file(s) and
 *  reports on all identifiers (constants, types, variables and
 *  procedure IDs) of the exporting DEF module.  It checks
 *  whether the IDs are being used by the importing module(s) or not
 *
 *  There are some special tricks with qualifiers:
 *
 *  /L   calculates the average identifier length
 *  /P   counts the number of procedures  (one file specification only!)
 *
 *  /T   generates test output
 *)

(*. FROM LogiFile IMPORT File, OpenMode, Open, Reset, Close; .*)
(*. FROM Terminal    IMPORT ReadString, Write, WriteString, WriteLn; .*)
(*. FROM CardinalIO  IMPORT WriteCardinal; .*)
(*. FROM Directories IMPORT DirQuery, DirResult; .*)
(*. FROM Strings     IMPORT CompareStr, Copy, Pos, Delete; .*)

(*.   FROM InOut IMPORT WriteCard,ReadString, Write, WriteString, WriteLn; .*)

  FROM SYSTEM IMPORT ADR, ADDRESS ;
   
  FROM DMFiles IMPORT
    Response, HiddenFileInfo, IOMode, TextFile, GetExistingFile, 
    CreateNewFile, Lookup, Close, Delete, Rename, Reset, Rewrite, 
    EOF, Again, ReadByte, WriteByte, EOL, ReadChar, SkipGap, 
    ReadChars, WriteChar, WriteEOL, WriteChars, legalNum, 
    GetCardinal, PutCardinal, GetInteger, PutInteger, GetReal, 
    PutReal, PutRealSci;        

  FROM DMStrings IMPORT String,  NewString,  (* GetString, *) 
    PutString,  DisposeString,  ConcatString, ConcatChar,  Length, 
    Concat,  ConcatCh,  AssignString,  LoadString, StoreString, 
    GetRString;  

  FROM DMMenus IMPORT Menu, Command, AccessStatus, Marking, 
    Separator, InstallAbout, InstallMenu, InstallCommand, 
    InstallAliasChar, InstallQuitCommand, DisableCommand, 
    EnableCommand, InstallSeparator, UncheckCommand, CheckCommand,
    ChangeCommandText, ChangeAliasChar; 

(*. 
  FROM DMWindows IMPORT Window, nonexistent, WindowKind, 
    ScrollBars, CloseAttr, ZoomAttr, WFFixPoint, WindowFrame, 
    CreateWindow, AutoRestoreProc, WindowExists, 
    RemoveWindow, NoBackground;                        
   .*)

  FROM DMWindowIO IMPORT CellHeight, SetPos, (*. WriteString, WriteLn, .*)
    (*. Write, .*) SetWindowFont, WindowFont, FontStyles, FontStyle,
    BackgroundWidth, BackgroundHeight, SelectForOutput, EraseContent;
        
  IMPORT DMWindowIO ;
  
  FROM DMMaster IMPORT MouseHandlers, InstallMouseHandler, 
    InstallSetUpProc, RunDialogMachine, DialogMachineTask;
    
  FROM DMEntryForms IMPORT FormFrame, WriteLabel, DefltUse, 
    CharField, StringField, CardField, IntField, RealField, 
    PushButton, RadioButtonID, DefineRadioButtonSet, RadioButton, 
    CheckBox, UseEntryForm;                            

  FROM DateAndTime IMPORT GetDateAndTime, DateAndTimeRec;
  
  FROM WriteDatTim IMPORT DateFormat, TimeFormat, WriteDate, WriteTime;

  FROM ClientsAux IMPORT 
    outF, outFOpen, writeOnFile, 
    Write, WriteString, WriteLn, WriteCardinal, WriteInteger, 
    CompareStr, killed ;
  
  FROM IDTables IMPORT ResetIDTables, MaxIDs, IDString, MaxIDLength, IDTablesSymbol,
    IDRecord,  ShowIDTable, ShowResIDTable,AddID, LookupID, CountUpForThisID,
    GetFirstIDRecord, GetNextIDRecord, WriteIDType, WriteLnAndCheck;

  FROM GetSym IMPORT
    InitGetSym, NextSymbol, LineNumber, GetCurrentID, copyFileToOutput;
    
  FROM AnalyzeProcs IMPORT
    CheckReferences, CountIDLengths, CountProcedures, defF, cliF ; 
  
  
(*.*********************************************************************.*)


  CONST
    MaxFilesInBatch = 64 ;
    FileNameLgth = 29 ; (*. MAC-filename can be maximum 30 chars long.*)
    
  TYPE
    FileName = ARRAY [0..FileNameLgth] OF CHAR ;
    FileType = (defExtension, otherExtension, fileNotFound) ;  
    BatchItem = RECORD
    			  fName: FileName ;
    			  fType: FileType
    			END;    			
    BatchItemList = ARRAY [1..MaxFilesInBatch] OF BatchItem ;
    
  VAR
    batchItem: BatchItemList ;
    
  TYPE
    DependencyVecType = 
    	ARRAY [1..MaxFilesInBatch] OF BOOLEAN ;
    DependencyMatType =
     ARRAY [1..MaxFilesInBatch] OF DependencyVecType ;
    	(*. element i,j is TRUE, if module in file j
    	imports (from) module in file i .*)
  VAR
    dMat: DependencyMatType ;

  TYPE
     MachineStates = (notReady, ready, running, halted);
    
  VAR
    curDMState: MachineStates;

    fileMenu, controlMenu, helpMenu
    : Menu ;
    
    chooseDefModuleCmd, chooseClientCmd,
    inputFromFileCmd, assignOutputCmd,
    setModeCmd,  runCmd,  haltResumeCmd, killCmd, helpCmd
    : Command ;

    inF
    : TextFile;
    
    defFOpen, cliFOpen, inFOpen,
    idTableChange,
    notOccuringIDsDisplay
    : BOOLEAN ;
    
(*.     CommandLine
    : ARRAY[0..FileNameLength] OF CHAR ;
    BlankPos, SlashPos
    : CARDINAL ;
 .*)    
    
    progHalted, doQuit
    : BOOLEAN ;
    

    mode:             (unusedReferencesMode, countProceduresMode, countIDLengthsMode) ;
    testOutput:       BOOLEAN ;     (* to turn on/off test output *)


  (**********************)
  (* About this program *)
  (**********************)

  PROCEDURE AboutProc;
  BEGIN
    SetPos(2,1);
    DMWindowIO.WriteString("     CrossRefs v 1.0");
    DMWindowIO.WriteLn;
    DMWindowIO.WriteString(" D. Keller and D.Gyalistras");
    DMWindowIO.WriteLn;
    DMWindowIO.WriteString("       20. Dez.1989"); 
  END AboutProc;
  

(*. *************************************************************** .*)
    
  PROCEDURE CloseFileAndUncheckCmd(VAR f: TextFile);
    VAR adr : ADDRESS ;
  BEGIN
    adr := ADR(f) ;
    IF adr = ADR(defF) THEN
      IF defFOpen THEN
  	    Close(defF); defFOpen:= FALSE;
        UncheckCommand(fileMenu, chooseDefModuleCmd) 
      END;(*IF*)
    ELSIF adr = ADR(cliF) THEN
      IF cliFOpen THEN
		Close(cliF); cliFOpen:= FALSE;
	    UncheckCommand(fileMenu, chooseClientCmd) 
	  END;(*IF*)
    ELSIF adr = ADR(inF) THEN
      IF inFOpen THEN
	    Close(inF); inFOpen:= FALSE;
	    UncheckCommand(fileMenu, inputFromFileCmd) 
	  END;(*IF*)
    ELSIF adr = ADR(outF) THEN
      IF outFOpen THEN
        Close(outF); outFOpen:= FALSE;
        UncheckCommand(fileMenu, assignOutputCmd) 
      END;(*IF*)
    END;(*IF*)
  END CloseFileAndUncheckCmd;
  
     
  PROCEDURE SetDMState(s: MachineStates);
  BEGIN
    CASE s OF
      notReady:
          IF killed OR (curDMState = notReady (*.start-up.*)) THEN
            CloseFileAndUncheckCmd (defF) ;
            CloseFileAndUncheckCmd (cliF) ;
            CloseFileAndUncheckCmd (inF) ;
            CloseFileAndUncheckCmd (outF) ;
            EnableCommand  (fileMenu, chooseDefModuleCmd);
            EnableCommand  (fileMenu, chooseClientCmd);
            EnableCommand  (fileMenu, inputFromFileCmd);
            EnableCommand  (fileMenu, assignOutputCmd);
            EnableCommand  (controlMenu, setModeCmd);
            EnableCommand  (controlMenu, runCmd);
            DisableCommand (controlMenu, haltResumeCmd);
            DisableCommand (controlMenu, killCmd);
            EnableCommand  (helpMenu, helpCmd);
          END;(*IF*)
    | ready:
          IF (curDMState = running) THEN
        	CloseFileAndUncheckCmd( outF ) ;
        	IF defFOpen THEN Close( defF ); Lookup( defF, defF.filename, FALSE ) END(*IF*);
        	IF cliFOpen THEN Close( cliF ); Lookup( cliF, cliF.filename, FALSE ) END(*IF*);
        	IF inFOpen  THEN Close( inF );  Lookup( inF, inF.filename, FALSE )   END(*IF*);
            EnableCommand  (fileMenu, chooseDefModuleCmd);
            EnableCommand  (fileMenu, chooseClientCmd);
            EnableCommand  (fileMenu, inputFromFileCmd);
            EnableCommand  (fileMenu, assignOutputCmd);
            EnableCommand  (controlMenu, setModeCmd);
            EnableCommand  (controlMenu, runCmd);
            DisableCommand (controlMenu, haltResumeCmd);
            DisableCommand (controlMenu, killCmd);
            EnableCommand  (helpMenu, helpCmd);
      	  END(*IF*);
    | running:
          killed := FALSE ;
          DisableCommand (fileMenu, chooseDefModuleCmd);
          DisableCommand (fileMenu, chooseClientCmd);
          DisableCommand (fileMenu, inputFromFileCmd);
          DisableCommand (fileMenu, assignOutputCmd);
          DisableCommand (controlMenu, setModeCmd);
          DisableCommand (controlMenu, runCmd);
          EnableCommand  (controlMenu, haltResumeCmd);
          EnableCommand  (controlMenu, killCmd);
          DisableCommand (helpMenu, helpCmd);
    	  ChangeCommandText (controlMenu, haltResumeCmd, "Halt"); 
    	  ChangeAliasChar   (controlMenu, haltResumeCmd, "H"); 
    | halted: 
          killed := FALSE ;
          DisableCommand (controlMenu, killCmd);
    	  ChangeCommandText (controlMenu, haltResumeCmd, "Resume"); 
    	  ChangeAliasChar   (controlMenu, haltResumeCmd, "R"); 
    END(*CASE*);
    curDMState:= s;
  END SetDMState;
  
  PROCEDURE ChooseDefModule ( );
  BEGIN
    idTableChange := TRUE ;
    CloseFileAndUncheckCmd (defF) ;
    GetExistingFile(defF, defF.filename);
    defFOpen := defF.res = done;
    IF defFOpen THEN
      CheckCommand(fileMenu, chooseDefModuleCmd);
      CloseFileAndUncheckCmd (inF) ;
      IF (cliFOpen AND (mode=unusedReferencesMode)) THEN 
        SetDMState(ready)
      END;(*IF*)
      WriteString("DEF-file: "); WriteString(defF.path); 
      WriteString(defF.filename); WriteLn;
    ELSE
      WriteString("DEF-file: "); Write("-"); WriteLn;    
      SetDMState(notReady) ;
    END;(*IF*)
  END ChooseDefModule ;
  
  PROCEDURE ChooseClient ( );
  BEGIN
    CloseFileAndUncheckCmd (cliF) ;
    GetExistingFile(cliF, cliF.filename);
    cliFOpen := cliF.res = done;
    IF cliFOpen THEN
      CheckCommand(fileMenu, chooseClientCmd);
      CloseFileAndUncheckCmd (inF) ;
      IF (defFOpen AND (mode=unusedReferencesMode)) OR 
         (mode=countProceduresMode) OR (mode=countIDLengthsMode) THEN
        SetDMState(ready) 
      END;(*IF*)
      WriteString("CLIENT-file: "); WriteString(cliF.path); 
      WriteString(cliF.filename); WriteLn;
    ELSE
      WriteString("CLIENT-file: "); Write("-"); WriteLn;    
      SetDMState(notReady) ;
    END;(*IF*)
  END ChooseClient ;
  
  PROCEDURE InputFromFile ( );
  BEGIN
    CloseFileAndUncheckCmd (inF) ;
    GetExistingFile(inF, inF.filename);
    inFOpen := inF.res = done;
    IF inFOpen THEN
      CheckCommand(fileMenu, inputFromFileCmd);
      CloseFileAndUncheckCmd (defF) ;
      CloseFileAndUncheckCmd (cliF) ;
      SetDMState(ready) ;
      WriteString("INPUT-file: "); WriteString(inF.path); 
      WriteString(inF.filename); WriteLn;
    ELSE
      WriteString("INPUT-file: "); Write("-"); WriteLn;    
      SetDMState(notReady) ;
    END;(*IF*)
  END InputFromFile ;
  
  PROCEDURE AssignOutput ( );
  BEGIN
    CloseFileAndUncheckCmd (outF) ;
    CreateNewFile(outF,"", outF.filename);
    outFOpen := outF.res = done;
    IF outFOpen THEN
      CheckCommand(fileMenu, assignOutputCmd);
      WriteString("OUTPUT-file: "); WriteString(outF.path); 
      WriteString(outF.filename); WriteLn;
    ELSE
      WriteString("OUTPUT-file: "); Write("-"); WriteLn;    
    END;(*IF*)
  END AssignOutput ;
  
  PROCEDURE QuitProc (VAR q: BOOLEAN);
  BEGIN
    CloseFileAndUncheckCmd (defF) ;
    CloseFileAndUncheckCmd (cliF) ;
    CloseFileAndUncheckCmd (inF) ;
    CloseFileAndUncheckCmd (outF) ;
    q := TRUE ;
    (*. HALT ; .*)
  END QuitProc ;
  
  PROCEDURE SetMode ( );
    CONST lm = 5; 
    VAR bf: FormFrame; notCancelled, ok: BOOLEAN; cl: INTEGER;
    VAR
      setModeRB,
      unusedReferencesRB,
      countProceduresRB,
      countIDLengthsRB: RadioButtonID;
  BEGIN
    cl:=2 ;
    WriteLabel ( cl, lm-4, "Program mode"); INC(cl);
    DefineRadioButtonSet(setModeRB);
    RadioButton( unusedReferencesRB, cl, lm+1,"unused References"); INC(cl); 
    CheckBox   ( cl, lm+6, "Display not occuring IDs", notOccuringIDsDisplay);  INC(cl);
    RadioButton( countIDLengthsRB,   cl, lm+1, "count ID-Lengths");  INC(cl);
    RadioButton( countProceduresRB,  cl, lm+1, "count Procedures"); INC(cl);
    WriteLabel ( cl, lm-4, "Output mode"); INC(cl);
    CheckBox   ( cl, lm+1, "Write test-output", testOutput); INC(cl); 
    INC(cl);INC(cl);

    CASE mode OF
    | countIDLengthsMode:
        setModeRB := countIDLengthsRB ;
    | countProceduresMode:
        setModeRB := countProceduresRB ;
    | unusedReferencesMode:
        setModeRB := unusedReferencesRB ;
    END(*CASE*);
    bf.x:= 0; bf.y:= -1;
    bf.lines:= cl; bf.columns:= 50;
    UseEntryForm( bf, notCancelled) ;
    IF notCancelled THEN
    
      IF setModeRB=countIDLengthsRB THEN
        mode := countIDLengthsMode ;
        IF (cliFOpen OR inFOpen) THEN SetDMState(ready) END(*IF*);
        
      ELSIF setModeRB=countProceduresRB THEN      
        mode := countProceduresMode ;
        IF (cliFOpen OR inFOpen) THEN SetDMState(ready) END(*IF*);
        
      ELSIF setModeRB=unusedReferencesRB THEN
        mode := unusedReferencesMode ;
        IF ((defFOpen AND cliFOpen) OR inFOpen) THEN SetDMState(ready) END(*IF*);
      END(*IF*);
      
      IF mode<>unusedReferencesMode THEN
        DisableCommand (fileMenu, chooseDefModuleCmd)
      ELSE
        EnableCommand (fileMenu, chooseDefModuleCmd)
      END(*IF*);
      
    END(*IF*);
  END SetMode ;
  
  
  PROCEDURE Run ( );
    VAR
      d: DateAndTimeRec;
      importOccures: BOOLEAN ;
      numFilesInBatch, defFileNr, cliFileNr: CARDINAL ;
      path: ARRAY [0..63] OF CHAR ;
      lgthLongestFName: CARDINAL;
         
    PROCEDURE AssignName (VAR f: TextFile; path: ARRAY OF CHAR; name: FileName);
      VAR j: CARDINAL ;
    BEGIN
      f.filename[0] := 0C ;
      Concat(f.filename, path);
      Concat(f.filename, name);
    END AssignName ;
    
    PROCEDURE ReadFileNamesAndTest;
      VAR
        i,j: CARDINAL ;
        ch: CHAR ;
        tstF: TextFile ;
              
      PROCEDURE NameHasDefExtension ( name: FileName ): BOOLEAN;
      BEGIN
        FOR j:= 0 TO FileNameLgth DO
          IF name[j]="." THEN
            IF (j<FileNameLgth-2) AND (name[j+1]="D") AND 
               (name[j+2]="E") AND (name[j+3]="F") THEN
              RETURN (TRUE);
            END;(*IF*)
          END;(*IF*)
        END;(*FOR*)
        RETURN (FALSE)
      END NameHasDefExtension ;
      
      PROCEDURE AddModExtensionIfNeeded ( VAR name: FileName );
      BEGIN
        j:= Length(name) ;
        IF (j<=FileNameLgth-3) AND (j>0) THEN
          IF (name[j-1]=".") THEN
            name[ j ] := "M";
            name[j+1] := "O";
            name[j+2] := "D";
            name[j+3] := 0C; 
          END(*IF*);
        END(*IF*);
      END AddModExtensionIfNeeded ;
      
    BEGIN
      i:= 1 ;
      lgthLongestFName := 0 ; 
      WriteLn;
      WriteString("Filenames from INPUT-file: "); 
      WriteString(inF.path); WriteString(inF.filename); WriteLn;
      WriteLn;
      ReadChars( inF, path );
      tstF.res  := notdone ;
      tstF.curIOMode := reading ;
      WriteString("Path: "); WriteString(path); WriteLn;   
      WHILE NOT EOF(inF) AND (i<=MaxFilesInBatch) AND NOT killed DO
          DialogMachineTask ;
          ch:=" ";
          WHILE (NOT EOF(inF)) AND (ch<>EOL) DO
              ReadChar( inF, ch );
          END;(*WHILE*)
          ReadChars( inF, batchItem[i].fName );
          j:= Length(batchItem[i].fName) ;
          IF lgthLongestFName<j THEN lgthLongestFName := j END;(*IF*)
          AddModExtensionIfNeeded (batchItem[i].fName) ;
          IF NOT EOF(inF) THEN
              AssignName( tstF, path, batchItem[i].fName );
              Lookup(tstF, tstF.filename, FALSE);
              WriteCardinal(i,2); WriteString(". ");
              IF tstF.res=done THEN
                WriteString("found: ");
                Close(tstF);
                IF NameHasDefExtension(batchItem[i].fName) THEN
                  batchItem[i].fType := defExtension ;
                ELSE
                  batchItem[i].fType := otherExtension ;
                END;(*IF*)
              ELSE
                WriteString("  ?  : ");
                batchItem[i].fType := fileNotFound ;
              END(*IF*);
              WriteString(tstF.filename); WriteLn;
              INC(i);
          END(*IF*);
      END;(*WHILE*)
      numFilesInBatch := i-1;
    END ReadFileNamesAndTest;
    
    PROCEDURE AnnounceRunStart ;
    BEGIN 
      WriteLn; WriteLn;
      WriteString('Program Clients, V 1.0' ); WriteLn;
      (*.WriteDate(d, Write, letMonth);
      WriteString(', '); 
      WriteTime(d, Write, brief24hSecs); WriteLn; WriteLn;.*)      
      WriteString('Mode: ' );
      CASE mode OF
      | countIDLengthsMode:
          WriteString( 'count ID-Lengths' ); 
      | countProceduresMode:
          WriteString( 'count Procedures' );
      | unusedReferencesMode:
          WriteString( 'unused References' );
      END(*CASE*);
      WriteLn;
      IF outFOpen THEN
        WriteString('Output on file: ' ); WriteString(outF.path);
        WriteString(outF.filename); WriteLn;
      END;(*IF*)              
      IF testOutput THEN 
        WriteString('Test-output switched on' ); WriteLn; WriteLn;
        ShowResIDTable;
        WriteLn;
      END;(*IF*)
    END AnnounceRunStart ;
    
    PROCEDURE AnnounceRunEnd ;
    BEGIN 
      WriteLn;
      WriteString('***** Ready' ); WriteLn;      
    END AnnounceRunEnd ;
        
    PROCEDURE DoOneClient ;
    BEGIN
      WriteLn; 
      IF inFOpen THEN 
        WriteString("*************************     DEF"); WriteCardinal(defFileNr,3);
        WriteString(" | CLI"); WriteCardinal(cliFileNr,3);
        WriteString("     *************************"); WriteLn;
      ELSE
        WriteString("***************************************************************************");
      END;(*IF*)
      WriteLn;WriteLn;
      CASE mode OF
      | countIDLengthsMode:
          ResetIDTables ;
          CountIDLengths ;
          idTableChange := TRUE ;
      | countProceduresMode:
          ResetIDTables ;
          CountProcedures ;
          idTableChange := TRUE ;
      | unusedReferencesMode:
          IF idTableChange THEN ResetIDTables END;(*IF*)
          CheckReferences( idTableChange, notOccuringIDsDisplay, importOccures );
          idTableChange := FALSE ;
      END(*CASE*);
    END DoOneClient ;
    
    PROCEDURE ProcessBatch ;
    
      PROCEDURE ResetDependencyMatrix ;
        VAR i,j: CARDINAL ;
      BEGIN
        FOR i:= 1 TO MaxFilesInBatch DO
          FOR j:= 1 TO MaxFilesInBatch DO
            dMat[i,j] := FALSE ;
          END(*FOR*)
        END(*FOR*);
      END ResetDependencyMatrix ;
      
      PROCEDURE ShowDependencyMatrix ;
        VAR i,j, nameLgth: CARDINAL ;
        CONST TAB = 11C ;

      BEGIN
        WriteLn; 
        WriteString("***************************************************************************");
        WriteLn;WriteLn;
        FOR i:=1 TO lgthLongestFName+5 DO Write(" ") END(*FOR*);
        FOR i:=1 TO numFilesInBatch DO
          Write(TAB); WriteCardinal(i,2)
        END(*FOR*); 
        WriteLn;     
        FOR i:= 1 TO numFilesInBatch DO
          WriteCardinal(i,2); WriteString(". ");
          nameLgth := Length(batchItem[ i ].fName) ;
          WriteString( batchItem[ i ].fName );
          WHILE nameLgth<lgthLongestFName+1 DO
            Write(" "); INC(nameLgth);
          END;(*WHILE*)
          FOR j:= 1 TO numFilesInBatch DO
            Write(TAB);
            IF dMat[i,j] THEN
              WriteString(" x")
            ELSE
              WriteString(" -");
            END;(*IF*)
          END;(*FOR*)
          WriteLn;
        END(*FOR*);
        WriteLn;
        WriteLn; 
      END ShowDependencyMatrix ;
      
    BEGIN
      ReadFileNamesAndTest ;
      ResetDependencyMatrix ;
      IF (mode = unusedReferencesMode) THEN
        defFileNr := 1 ;
      ELSE
        (*. enter outer loop only once .*)
        defFileNr := numFilesInBatch ;
      END;(*IF*)
      WHILE (defFileNr <= numFilesInBatch) AND NOT killed DO
          IF ( (mode = unusedReferencesMode)
          AND (batchItem[ defFileNr ].fType = defExtension ) ) 
          OR (mode <> unusedReferencesMode) THEN
              AssignName( defF, path, batchItem[ defFileNr ].fName );
              Lookup( defF, defF.filename, FALSE ) ;
              idTableChange := TRUE ;
              cliFileNr := 1 ;
              WHILE (cliFileNr <= numFilesInBatch) AND NOT killed DO
                  IF ( (mode = unusedReferencesMode)
                  AND (batchItem[ cliFileNr ].fType = otherExtension ) )
                  OR (mode <> unusedReferencesMode) THEN
                      AssignName( cliF, path, batchItem[ cliFileNr ].fName );
                      Lookup( cliF, cliF.filename, FALSE );
                      DoOneClient ;
                      Close( cliF );
                      dMat[ defFileNr, cliFileNr ] := importOccures ;
                  END;(*IF*)
                  INC(cliFileNr);             
              END;(*WHILE*)                  
              Close( defF );
          END;(*IF*)
          INC(defFileNr);
      END;(*WHILE*)
      IF (mode = unusedReferencesMode) AND NOT killed THEN
        ShowDependencyMatrix ;
      END;(*IF*)
    END ProcessBatch ;

  BEGIN
    IF (curDMState = ready) THEN
      SetDMState(running) ;
      AnnounceRunStart ;
      writeOnFile      := outFOpen ;
      copyFileToOutput := testOutput ;            
      IF inFOpen THEN ProcessBatch ELSE DoOneClient END;(*IF*)              
      IF NOT killed THEN SetDMState(ready) END(*IF*);
      copyFileToOutput := FALSE ;      
      writeOnFile      := FALSE ;
      AnnounceRunEnd ;
    ELSE (*. curDMState<> ready .*)
      writeOnFile := FALSE ;
      WriteString('>>>  Can not run. Preconditions for starting a run:' ); WriteLn;
      WriteString('>>>  unused References Mode:   INPUT-file open, or DEF- and CLIENT-file open' ); WriteLn;
      WriteString('>>>  count Procedures Mode :   INPUT- or CLIENT-file open' ); WriteLn;
      WriteString('>>>  count ID-LengthsMode  :   INPUT- or CLIENT-file open' ); WriteLn;
      WriteLn;
    END(*IF*);
  END Run ;
  
  PROCEDURE Halt ( );
  BEGIN
    IF (curDMState = running) THEN
      progHalted:=TRUE ;
      SetDMState(halted) ;
      WriteLn; WriteLn;
      WriteString('***** Program halted... '); 
      REPEAT
        DialogMachineTask
      UNTIL NOT progHalted ;
    ELSIF (curDMState = halted) THEN
      progHalted := FALSE ;
      SetDMState(running) ;
      WriteString('resumed'); WriteLn; WriteLn;
    END(*IF*);   
  END Halt ;
  
  PROCEDURE Kill ( );
  BEGIN
    IF (curDMState = running) THEN
      killed:=TRUE ;
      SetDMState(notReady) ;
      WriteLn;
      WriteString('***** Program killed'); WriteLn;
    END(*IF*);   
  END Kill ;
  
  PROCEDURE Help ( );
  BEGIN
    WriteLn;
    WriteString(' ================================ HELP ================================' ); WriteLn;
    WriteString(' The program determines which identifiers declared in a DEF-module' ); WriteLn;
    WriteString(' are imported (and used) in other modules, and which not.' ); WriteLn;
    WriteLn;
    WriteString(' Menu-Commands:' ); WriteLn;
    WriteLn;
    WriteString(' "File/Choose DEF-module...":' ); WriteLn;
    WriteString('    Select this command to determine the DEF-Module which exports the' ); WriteLn;
    WriteString('    idents to be searched in the client-module.' ); WriteLn;
    WriteString(' "File/Choose CLIENT...":' ); WriteLn;
    WriteString('    Select this command to determine the file containing the client-module.' ); WriteLn;
    WriteString(' "File/Input from file...":' ); WriteLn;
    WriteString('    With this command you can reassign input to a ASCII-file. All files' ); WriteLn;
    WriteString('    with extension ".DEF" mentioned in the INPUT-file (typically a MAKE-' ); WriteLn;
    WriteString('    file) are taken as DEF-modules. Their idents are searched in all' ); WriteLn;
    WriteString('    other files mentioned in the INPUT-file.' ); WriteLn;
    WriteString('    Selecting a file by this command will override any selections done' ); WriteLn;
    WriteString('    by means of the commands "Choose DEF-module..." and "Choose Client..."' ); WriteLn;
    WriteString(' "File/Output to file...":' ); WriteLn;
    WriteString('    Allows You to select a file on which all output will be dumped.' ); WriteLn;
    WriteString(' "File/Quit...":' ); WriteLn;
    WriteString('    Quit this program.' ); WriteLn;
    WriteLn;
    WriteString(' "Control/Set Mode...":' ); WriteLn;
    WriteString('    Displays a box allowing You to change settings that determine output-' ); WriteLn;
    WriteString('    and analysis-modes.' ); WriteLn;
    WriteString(' "Control/Run...":' ); WriteLn;
    WriteString('    Starts analysis of client-module(s), if a DEF-module and a client-' ); WriteLn;
    WriteString('    module or an INPUT-file have (has) been chosen.' ); WriteLn;
    WriteString(' "Control/Halt (Resume)...":' ); WriteLn;
    WriteString('    Halts a running analysis. Continue with "Control/Resume..."' ); WriteLn;
    WriteString(' "Control/Kill...":' ); WriteLn;
    WriteString('    Allows You to abort a running analysis at any time. ' ); WriteLn;
    WriteLn;
    WriteString(' "Help/Display Help...":' ); WriteLn;
    WriteString('    Writes this text on the screen.' ); WriteLn;
    WriteLn;
	WriteString(' ======================================================================' ); 
  END Help ;
  
  
  (*.********************************************************************  .*)
  
  PROCEDURE SettingUp;
  BEGIN
    (*jrnlW    := nonexistent;*)
    killed:=FALSE;
    defFOpen := FALSE; defF.res := notdone; defF.filename := "DEF-file";
    cliFOpen := FALSE; cliF.res := notdone; cliF.filename := "CLIENT-file";
    inFOpen  := FALSE; inF.res  := notdone; inF.filename  := "INPUT-file";
    outFOpen := FALSE; outF.res := notdone; outF.filename := "OUTPUT-file";
    SetDMState(notReady);
    WriteLn; WriteLn;
  END SettingUp;
  
   
  (*********************************************)
  (* Initialization of DM before it is running *)
  (*********************************************)
    
  PROCEDURE InstallMenus;
  BEGIN
    InstallAbout("About CrossRefs ...",190,80,AboutProc);
    
    InstallMenu(fileMenu,'File',enabled);        
    InstallCommand( fileMenu, chooseDefModuleCmd, "Choose DEF-file...",
    				ChooseDefModule, enabled, unchecked);
    InstallAliasChar(fileMenu, chooseDefModuleCmd, "D");        
    InstallCommand(	fileMenu, chooseClientCmd, "Choose CLIENT-file...",
    				ChooseClient, enabled, unchecked);
    InstallAliasChar(fileMenu, chooseClientCmd, "C");    
    InstallSeparator(fileMenu, line);       
    InstallCommand(	fileMenu, inputFromFileCmd, "Input from file...",
    				InputFromFile, disabled, unchecked);
    InstallAliasChar(fileMenu, inputFromFileCmd, "I");    
    InstallCommand( fileMenu, assignOutputCmd,"Output to file...",
    				AssignOutput, enabled, unchecked);
    InstallAliasChar(fileMenu, assignOutputCmd, "O");                   
    InstallQuitCommand("Quit",QuitProc(doQuit),"Q");

    InstallMenu(controlMenu,'Control',enabled);        
    InstallCommand( controlMenu, setModeCmd, "Set Mode...",
    				SetMode, enabled, unchecked);
    InstallAliasChar(controlMenu, setModeCmd, "M");        
    InstallCommand( controlMenu, runCmd, "Run",
    				Run, enabled, unchecked);
    InstallAliasChar(controlMenu, runCmd, "R");        
    InstallSeparator(controlMenu, line);                      
    InstallCommand( controlMenu, haltResumeCmd, "Halt",
    				Halt, enabled, unchecked);
    InstallAliasChar(controlMenu, killCmd, "H"); 
    InstallCommand( controlMenu, killCmd, "Kill",
    				Kill, enabled, unchecked);
    InstallAliasChar(controlMenu, killCmd, "K"); 
           
    InstallMenu(helpMenu,'Help',enabled);        
    InstallCommand( helpMenu, helpCmd, "Display Help...",
    				Help, enabled, unchecked);
    InstallAliasChar(helpMenu, helpCmd, "E"); 
   
    InstallSetUpProc(SettingUp);
  END InstallMenus;
  
BEGIN
  curDMState := notReady ;
  mode := unusedReferencesMode ;
  testOutput := FALSE ;
  idTableChange := TRUE ;
  notOccuringIDsDisplay := FALSE ;
  InstallMenus;
  RunDialogMachine
END Clients .
