IMPLEMENTATION MODULE GetSym;
(*
 *  Reads the next symbol from a MODULA source file
 *)

  FROM DMMaster IMPORT DialogMachineTask;

  FROM DMFiles IMPORT
    TextFile, ReadChar, EOL, EOF(*. new! .*); 
  
  FROM ClientsAux IMPORT 
    Write, WriteString, WriteLn, 
    WriteCardinal, WriteInteger, CompareStr, killed;
  
  FROM IDTables    IMPORT
    IDTablesSymbol, IDString, EmptyID, MaxIDLength,
    LookupReservedID, WriteLnAndCheck;

VAR
   eof:           BOOLEAN;      (* new !!!! *) 
   InFile:        TextFile;     (* new !!!! *)
   ch:            CHAR;
   CurrentID:     IDString;
   LineNo:        CARDINAL;     (* line number in input file *)
   CommentLevel:  CARDINAL;     (* tracking nested comments *)
   ScreenCopying: BOOLEAN;


PROCEDURE InitGetSym( FileToReadFrom: TextFile );
BEGIN
   InFile := FileToReadFrom;
   LineNo := 1;
   ch := " ";
   eof := FALSE;
END InitGetSym;


PROCEDURE GetCurrentID( VAR ID: IDString );
BEGIN
   ID := CurrentID;
END GetCurrentID;


PROCEDURE LineNumber(): CARDINAL;
BEGIN
   RETURN LineNo;
END LineNumber;


PROCEDURE Get( VAR c: CHAR );
BEGIN
   IF (NOT EOF( InFile ) AND (NOT killed) ) THEN
      ReadChar( InFile, c );
      IF c = EOL THEN
         c := " ";
         INC( LineNo );
         IF copyFileToOutput THEN
            WriteLnAndCheck;
         END;
      ELSIF copyFileToOutput THEN
         Write( c );
      END;
      eof := FALSE;
   ELSE
      eof := TRUE;
   END;
END Get;


PROCEDURE NextSymbol(): IDTablesSymbol;
VAR
   TerminatingChar:  CHAR;     (* for strings, either " or ' *)
   CurrentIDLength:  CARDINAL;
   CurrentSymbol:    IDTablesSymbol;
BEGIN
   REPEAT
      DialogMachineTask;
      WHILE (ch = " ") AND NOT killed DO
         Get( ch );
      END;
      IF ((CAP(ch) >= "A") AND (CAP(ch) <= "Z")) THEN
         CurrentID:= EmptyID; CurrentIDLength:= 0;
         REPEAT
            IF CurrentIDLength < MaxIDLength THEN
               CurrentID[ CurrentIDLength ]:= ch;
               INC( CurrentIDLength );
            END;
            Get( ch )
         UNTIL NOT ( ((CAP(ch) >= "A") AND (CAP(ch) <= "Z"))
                     OR ((ch >= "0") AND (ch <= "9")));
         LookupReservedID( CurrentID, CurrentSymbol );
      ELSIF ch = "(" THEN
         Get( ch );
         IF ch = "*" THEN
            INC( CommentLevel );
            REPEAT
               REPEAT
                  Get( ch );
                  (* catch nested comments *)
                  IF ch = "(" THEN
                     CurrentSymbol := NextSymbol();
                  END;
               UNTIL (ch = "*") OR (eof);
               WHILE ch = "*" DO
                  Get( ch );
               END;
            UNTIL (ch = ")") OR (eof);
            DEC( CommentLevel );
            IF eof THEN
               CurrentSymbol := EndSymbol;
            ELSE
               CurrentSymbol := Comment;
               Get( ch );
            END;
         ELSE
            CurrentSymbol := OpenParen;
         END;
      ELSIF ch = ")" THEN
         CurrentSymbol := CloseParen;
         Get( ch );
      ELSIF ch = ";" THEN
         CurrentSymbol := Semicolon;
         Get( ch );
      ELSIF ch = "," THEN
         CurrentSymbol := Comma;
         Get( ch );
      ELSIF ch = "." THEN
         CurrentSymbol := Dot;
         Get( ch );
      ELSIF ch = ":" THEN
         Get( ch );
         IF ch = "=" THEN
            CurrentSymbol := Other;
            Get( ch );
         ELSE
            CurrentSymbol := Colon;
         END;
      ELSIF (ch = "<") OR (ch = ">") THEN
         Get( ch );
         IF ch = "=" THEN
            Get( ch );
            CurrentSymbol := Other;
         ELSE
            CurrentSymbol := Other;
         END;
      ELSIF ch = "=" THEN
         CurrentSymbol := EqualSymbol;
         Get( ch );
      ELSIF (ch = "'") OR (ch = '"') THEN
         TerminatingChar := ch;
         REPEAT
            Get( ch );
         UNTIL (ch = TerminatingChar) OR (eof);
         IF ch = TerminatingChar THEN
            CurrentSymbol := StringSym;
            Get( ch );
         ELSE
            CurrentSymbol := EndSymbol;
         END;
      ELSIF eof THEN
         CurrentSymbol := EndSymbol;
      ELSE
         CurrentSymbol := Other;
         Get( ch );
      END;
   UNTIL (CurrentSymbol <> Comment) OR (CommentLevel > 0) OR killed;
   RETURN CurrentSymbol;
END NextSymbol;


BEGIN
   copyFileToOutput := FALSE;
END GetSym.
