IMPLEMENTATION MODULE IDTables;
(*
 *  Implements tables for the storage of identifiers
 *  and reserved words.
 *)

  FROM DMMaster IMPORT DialogMachineTask;
  
  FROM ClientsAux IMPORT 
    Write, WriteString, WriteLn, WriteCardinal, WriteInteger,
    CompareStr, Read, BusyRead, killed ;

CONST
   MaxResIDs    =  40;   (* Max # of reserved words (in Modula: 40) *)


VAR
   IDTable:       ARRAY [1..MaxIDs] OF IDRecord; (* identifiers *)
   ResIDTable:    ARRAY [1..MaxResIDs] OF RECORD (* reserved words *)
                        					  String:  IDString;
                       					 	  WhichID: IDTablesSymbol;
                     					  END;
   IDCount:       CARDINAL;
   ResIDCount:    CARDINAL;
   NextIDIndex:   CARDINAL;
   
   
   
PROCEDURE KeyPressed(): BOOLEAN;
  VAR ch : CHAR;
BEGIN
  BusyRead(ch);
  RETURN ch=0C;
END KeyPressed;   

PROCEDURE AddID( ID: IDString; PlaceToInsert: CARDINAL; IDType: IDTablesSymbol;
                               LineNumber: CARDINAL;  VAR OK: BOOLEAN );
VAR i: CARDINAL;
BEGIN
   IF IDCount < MaxIDs THEN
      (* first shift some up by one to make room *)
      FOR i:= IDCount TO PlaceToInsert BY -1 DO
         IDTable[i+1]:= IDTable[i];
      END;
      (* now insert *)
      IDTable[ PlaceToInsert ].String := ID;
      IDTable[ PlaceToInsert ].LineNo := LineNumber;
      IDTable[ PlaceToInsert ].TypeOfID := IDType;
      IDTable[ PlaceToInsert ].imported := FALSE ;
      IDTable[ PlaceToInsert ].qualifUse := FALSE ;
      IDTable[ PlaceToInsert ].strOccur := 0 ;
      INC( IDCount );
      OK := TRUE;
   ELSE
      OK := FALSE;
   END;
END AddID;


PROCEDURE LookupID( ID: IDString; VAR Position: CARDINAL; VAR Found: BOOLEAN );
VAR
   Upper, Lower, Middle: CARDINAL; (* indices for the binary search *)
BEGIN
   Lower:= 1; Upper:= IDCount;
   WHILE Lower <= Upper DO
      Middle:= (Lower+Upper) DIV 2;
      IF CompareStr( IDTable[ Middle ].String, ID) <= 0 THEN
         Lower:= Middle+1;
      END;
      IF CompareStr( IDTable[ Middle ].String, ID) >= 0 THEN
         Upper:= Middle-1;
      END;
   END;
   Found := Lower-Upper = 2;
   Position := Upper + 1;
END LookupID;
 
  
PROCEDURE CountUpForThisID( ID: IDString; occur: OccurenceType ;
							VAR OK: BOOLEAN );
VAR
   Position: CARDINAL;
   found:    BOOLEAN;
BEGIN
   LookupID( ID, Position, found);
   IF found THEN
      INC( IDTable[ Position ].strOccur );
      CASE occur OF
        inImportList     : IDTable[ Position ].imported := TRUE ;
      | usedQualified    : IDTable[ Position ].qualifUse := TRUE ;
      | otherOccurence   : ;
      END(*CASE*);
      OK := TRUE;
   ELSE
      OK := FALSE;
   END;
END CountUpForThisID;


PROCEDURE LookupReservedID( ID: IDString; VAR WhichOne: IDTablesSymbol );
VAR
   Upper, Lower, Middle: CARDINAL;
BEGIN
   (* binary search in the ReservedWordID table *)
   Lower:= 1; Upper:= ResIDCount;
   WHILE Lower <= Upper DO
      Middle:= (Lower+Upper) DIV 2;
      IF CompareStr( ResIDTable[ Middle ].String, ID) <= 0 THEN
         Lower:= Middle+1;
      END;
      IF CompareStr( ResIDTable[ Middle ].String, ID) >= 0 THEN
         Upper:= Middle-1;
      END;
   END;
   IF Lower-Upper = 2 THEN (* we have found it in the table *)
      WhichOne := ResIDTable[ Upper+1 ].WhichID;
   ELSE
      WhichOne := Identifier;
   END;
END LookupReservedID;


PROCEDURE WriteIDType( IDType: IDTablesSymbol );
BEGIN
   CASE IDType OF
      ConstID:     WriteString( "constant " ); |
      TypeID:      WriteString( "type     " ); |
      VarID:       WriteString( "variable " ); |
      ProcedureID: WriteString( "procedure" );
   ELSE
      WriteString( "?? unknown ID type ??" );
   END;
END WriteIDType;


PROCEDURE WriteLnAndCheck;
VAR
   ch: CHAR;
BEGIN
   (* just write an ordinary "WriteLn" *)
   WriteLn;
   (* but then check if any key has been pressed *)
(*.    IF KeyPressed() THEN
      REPEAT
         Read( ch );
      UNTIL NOT (KeyPressed() OR killed) ;
      (* now stop and wait for another keystroke, after that continue *)
      Read( ch );
   END;
 .*)
END WriteLnAndCheck;


PROCEDURE ShowResIDTable;
VAR i: CARDINAL;
BEGIN
   WriteLn;
   WriteString( "Tabelle der reservierten Bezeichner" ); WriteLn;
   FOR i := 1 TO ResIDCount DO
      IF NOT killed THEN
        WriteCardinal( i, 4 ); WriteString(":  ");
        WriteString( ResIDTable[ i ].String );
        WriteLnAndCheck;
        DialogMachineTask;
      END;
   END;
END ShowResIDTable;


PROCEDURE ShowIDTable;
VAR i: CARDINAL;
BEGIN
   WriteLn;
   WriteString( "Tabelle aller eingelesenen Bezeichner" ); WriteLn;
   FOR i := 1 TO IDCount DO
      IF NOT killed THEN
        WriteCardinal( i, 4 ); WriteString(":  ");
        WriteString( IDTable[ i ].String ); WriteString( "  " );
        WriteIDType( IDTable[ i ].TypeOfID );
        WriteLnAndCheck;
        DialogMachineTask;
      END;
   END;
END ShowIDTable;

PROCEDURE GetNextIDRecord(  VAR IDR: IDRecord; VAR done: BOOLEAN );
BEGIN
   IF NextIDIndex <= IDCount THEN
      IDR := IDTable[ NextIDIndex ];
      INC( NextIDIndex );
      done := TRUE;
   ELSE
      done := FALSE;
   END;
END GetNextIDRecord;


PROCEDURE GetFirstIDRecord( VAR IDR: IDRecord; VAR done: BOOLEAN );
BEGIN
   NextIDIndex := 1;
   GetNextIDRecord( IDR, done );
END GetFirstIDRecord;

 
PROCEDURE AddReservedID( NewID: ARRAY OF CHAR; WhichSymbol: IDTablesSymbol );
VAR
   ID:                   IDString;
   i:                    CARDINAL;
   Upper, Lower, Middle: CARDINAL;
   PlaceToInsert:        CARDINAL;
BEGIN
   IF ResIDCount < MaxResIDs THEN
      ID := EmptyID;
      FOR i := 0 TO HIGH( NewID ) DO
         ID[ i ] := NewID[ i ];
      END;
      (* binary search in the ReservedWordID table *)
      Lower:= 1; Upper:= ResIDCount;
      WHILE Lower <= Upper DO
         Middle:= (Lower+Upper) DIV 2;
         IF CompareStr( ResIDTable[ Middle ].String, ID) <= 0 THEN
            Lower:= Middle+1;
         END;
         IF CompareStr( ResIDTable[ Middle ].String, ID) >= 0 THEN
            Upper:= Middle-1;
         END;
      END;
      IF Lower-Upper < 2 THEN (* it is not yet in the table *)
         PlaceToInsert := Upper+1;
         (* first shift some up by one to make room *)
         FOR i:= ResIDCount TO PlaceToInsert BY -1 DO
            ResIDTable[i+1]:= ResIDTable[i];
         END;
         (* now insert *)
         ResIDTable[ PlaceToInsert ].String := ID;
         ResIDTable[ PlaceToInsert ].WhichID := WhichSymbol;
         INC( ResIDCount );
      END;
   END;
END AddReservedID;

  PROCEDURE ResetIDCountsOfAllRecords ;
    VAR i: CARDINAL ;
  BEGIN
    i:=1 ;
    WHILE (i <= IDCount) DO 
      IDTable[ i ].strOccur := 0 ;       
      IDTable[ i ].imported := FALSE ;       
      IDTable[ i ].qualifUse := FALSE ;       
      INC(i) 
    END(*WHILE*);
  END ResetIDCountsOfAllRecords;
  
  PROCEDURE ResetIDTables ;
  BEGIN
    IDCount := 0 ;
    NextIDIndex := 1 ;
  END ResetIDTables;
  

BEGIN
   IDCount := 0;
   ResIDCount := 0;
   NextIDIndex := 1;

   AddReservedID("AND", Other);          AddReservedID("ARRAY", Other);          
   AddReservedID("BEGIN", Other);        AddReservedID("BY", Other);             
   AddReservedID("CASE", Other);         AddReservedID("CONST", ConstID);          
   AddReservedID("DEFINITION", DefinitionID);
   AddReservedID("DIV", Other);            
   AddReservedID("DO", Other);           AddReservedID("ELSE", Other);           
   AddReservedID("ELSIF", Other);        AddReservedID("END", Other);            
   AddReservedID("EXIT", Other);         AddReservedID("EXPORT", Other);         
   AddReservedID("FOR", Other);          AddReservedID("FROM", Other);           
   AddReservedID("IF", Other);           
   AddReservedID("IMPLEMENTATION", ImplementationID);
   AddReservedID("IMPORT", ImportID);    AddReservedID("IN", Other);             
   AddReservedID("LOOP", Other);         AddReservedID("MOD", Other);
   AddReservedID("MODULE", ModuleID);    AddReservedID("NOT", Other);
   AddReservedID("OF", Other);           AddReservedID("OR", Other);
   AddReservedID("POINTER", Other);      
   AddReservedID("PROCEDURE", ProcedureID);
   AddReservedID("QUALIFIED", Other);    AddReservedID("RECORD", Other);
   AddReservedID("REPEAT", Other);       AddReservedID("RETURN", Other);
   AddReservedID("SET", Other);          AddReservedID("THEN", Other);
   AddReservedID("TO", Other);           AddReservedID("TYPE", TypeID);
   AddReservedID("UNTIL", Other);        AddReservedID("VAR", VarID);
   AddReservedID("WHILE", Other);        AddReservedID("WITH", Other);

END IDTables.
